@@ -1,4 +1,6 @@
-/* Copyright (c) 2003, 2004 SuSE Linux AG, Germany
+/*
+ * Copyright (c) 2005, 2006, 2007, 2008, 2009 SUSE LINUX Products GmbH, Germany
+ * Copyright (c) 2003, 2004 SuSE Linux AG, Germany
  * Copyright (c) 1999, 2000, 2001, 2002 SuSE GmbH Nuernberg, Germany
  * Author: Thorsten Kukuk <kukuk@suse.de>
  *
@@ -28,100 +30,110 @@
  * login -h hostname    (for telnetd, etc.)
  * login -f name        (for pre-authenticated login: datakit, xterm, etc.)
  * login -p             (don't destroy environment)
+ * login -H             (suppress printing hostname)
  */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#define _GNU_SOURCE
-
-#include <sys/param.h>
-#include <stdio.h>
-#include <ctype.h>
-#include <unistd.h>
+#include <errno.h>
+#include <fcntl.h>
 #include <getopt.h>
-#include <memory.h>
-#include <time.h>
-#include <sys/stat.h>
-#include <sys/time.h>
-#include <sys/resource.h>
-#include <sys/file.h>
-#include <termios.h>
+#include <grp.h>
+#include <locale.h>
+#include <netdb.h>
+#include <pwd.h>
+#include <setjmp.h>
+#include <signal.h>
+#include <stdio.h>
 #include <string.h>
 #include <sys/ioctl.h>
+#include <sys/param.h>
+#include <sys/resource.h>
 #include <sys/wait.h>
-#include <signal.h>
-#include <errno.h>
-#include <grp.h>
-#include <pwd.h>
+#include <termios.h>
 #include <utmp.h>
-#include <setjmp.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/syslog.h>
-#include <sys/sysmacros.h>
-#include <netdb.h>
-#include <locale.h>
-#include <libintl.h>
-#include <sys/sysmacros.h>
-#include <linux/major.h>
+#include <syslog.h>
 #include <security/pam_appl.h>
 #include <security/pam_misc.h>
-
-#ifdef WITH_SELINUX
-#include <selinux/get_context_list.h>
-#include <selinux/flask.h>
-#include <selinux/selinux.h>
+#ifdef WITH_AUDIT
+#include "libaudit.h"
 #endif
 
-#include "faillog.h"
-#include "failure.h"
-
 #define PAM_FAIL_CHECK if (retcode != PAM_SUCCESS) { \
        fprintf (stderr, "\n%s\n", pam_strerror (pamh, retcode)); \
        syslog (LOG_ERR, "%s", pam_strerror (pamh, retcode)); \
-       sleep (getdef_num ("FAIL_DELAY", 1)); \
+       sleep (getlogindefs_num ("FAIL_DELAY", 1)); \
        pam_end (pamh, retcode); \
        exit (1); \
    }
-#define PAM_END { pam_setcred(pamh, PAM_DELETE_CRED); \
-                  retcode = pam_close_session(pamh,0); \
+#define PAM_END { retcode = pam_close_session(pamh,0); \
+                  pam_setcred(pamh, PAM_DELETE_CRED); \
 		  pam_end(pamh,retcode); }
 
 #include "setproctitle.h"
-#include "getdef.h"
+#include "logindefs.h"
 
-#ifndef _
-#define _(s) gettext(s)
+#ifdef ENABLE_NLS
+#include <libintl.h>
+#define _(Msgid) gettext(Msgid)
+#else
+#define _(Msgid) ((const char *) (Msgid))
+#define textdomain(Domainname)
+#define bindtextdomain(Domainname, Dirname)
 #endif
 
-#ifdef WITH_SELINUX
-int selinux_enabled = -1;
-security_context_t user_context=NULL;
-security_context_t ttyn_context=NULL; /* The current context of ttyn device */
-security_context_t vcsn_context=NULL; /* The current context of vcsn device */
-security_context_t vcsan_context=NULL;/* The current context of vcsan device */
+#ifdef WITH_DEBUG
+#include <stdarg.h>
+
+static int do_debug = 0;
+
+static void
+debug (const char *fmt,...)
+{
+  va_list ap;
+  char msg[512];
+
+  if (do_debug == 0)
+    return;
+
+  va_start (ap, fmt);
+  vsnprintf (msg, sizeof (msg), fmt, ap);
+
+  if (do_debug == 1)
+    {
+      fputs (msg, stderr);
+      fputs ("\n", stderr);
+      fflush (stderr);
+    }
+  else
+    syslog (LOG_DEBUG, "%s", msg);
+
+  va_end (ap);
+}
 #endif
 
 /* Print the version information.  */
-static inline void
+static void
 print_version (void)
 {
   fprintf (stdout, "login (%s) %s\n", PACKAGE, VERSION);
-  fprintf (stdout, gettext ("\
+  fprintf (stdout, _("\
 Copyright (C) 1980, 1987, 1988 The Regents of the University of California.\n\
 Copyright (C) 1999, 2000, 2001, 2002 SuSE GmbH Nuernberg, Germany.\n\
+Copyright (C) 2003, 2004 SuSE Linux AG Nuernberg, Germany.\n\
+Copyright (C) 2005, 2006, 2007 SUSE LINUX Products GmbH Nuernberg, Germany.\n\
 This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 "));
   fprintf (stdout, _("Written by Thorsten Kukuk.\n"));
 }
 
-static inline void
+static void
 print_usage (void)
 {
-  fputs (_("Usage: login [name] [-p] [-h host] [-f name]\n"),
+  fputs (_("Usage: login [-p] [-h host] [-f name|name]\n"),
 	 stdout);
 }
 
@@ -135,12 +147,14 @@ print_help (void)
   fputs (_("  -h hostname    Name of the remote host for umtp/wtmp\n"),
 	 stdout);
   fputs (_("  -f name        Skip a second login authentication\n"), stdout);
+  fputs (_("  -H             Do not print hostname before login: prompt\n"),
+  	 stdout);
   fputs (_("  --help         Give this help list\n"), stdout);
   fputs (_("  --usage        Give a short usage message\n"), stdout);
   fputs (_("  --version      Print program version\n"), stdout);
 }
 
-static inline void
+static void
 print_error (void)
 {
   const char *program = "login";
@@ -172,15 +186,16 @@ motd (void)
   int fd, nchars;
   void (*oldint) (int);
   char tbuf[8192];
-  char motdlist[BUFSIZ], *mb, *motdfile;
+  char motdlist[BUFSIZ], *motdfile, *cp;
+  const char *mb;
 
-  if ((mb = getdef_str ("MOTD_FILE")) == NULL)
+  if ((mb = getlogindefs_str ("MOTD_FILE", NULL)) == NULL)
     return;
 
   strncpy (motdlist, mb, sizeof (motdlist));
   motdlist[sizeof (motdlist) - 1] = '\0';
 
-  for (mb = motdlist; (motdfile = strtok (mb, ":")) != NULL; mb = NULL)
+  for (cp = motdlist; (motdfile = strtok (cp, ":")) != NULL; cp = NULL)
     {
 
       if ((fd = open (motdfile, O_RDONLY, 0)) >= 0)
@@ -188,7 +203,8 @@ motd (void)
 	  oldint = signal (SIGINT, sigint);
 	  if (setjmp (motdinterrupt) == 0)
 	    while ((nchars = read (fd, tbuf, sizeof (tbuf))) > 0)
-	      write (fileno (stdout), tbuf, nchars);
+	      if (write (fileno (stdout), tbuf, nchars) == -1)
+		break;
 	  signal (SIGINT, oldint);
 	  close (fd);
 	}
@@ -196,6 +212,7 @@ motd (void)
 }
 
 static struct hostent *hostaddress = NULL;
+static struct hostent resulthostbuf;
 static int timeout;
 
 static void
@@ -212,53 +229,6 @@ timedout (int sig __attribute__ ((unused)))
   exit (0);
 }
 
-#ifdef WITH_SELINUX
-
-static security_context_t
-security_label_tty (const char *tty, security_context_t usercon)
-{
-  security_context_t newdev_context=NULL; /* The new context of a device */
-  security_context_t prev_context=NULL; /* The new context of a device */
-  if (getfilecon (tty, &prev_context) < 0)
-    {
-      fprintf (stderr, _("Warning!  Could not get current context for %s, not relabeling.\n"), tty);
-      syslog (LOG_NOTICE, "Warning!  Could not get current context for %s, notrelabeling.\n", tty);
-      return NULL;
-    }
-  if (security_compute_relabel (usercon, prev_context, SECCLASS_CHR_FILE,
-				&newdev_context) != 0)
-    {
-      fprintf (stderr, _("Warning!  Could not get new context for %s, not relabeling.\n"), tty);
-      syslog (LOG_NOTICE, "Warning!  Could not get new context for %s, not relabeling.\n", tty);
-      return NULL;
-    }
-  if (setfilecon (tty,newdev_context))
-    {
-      fprintf (stderr, _("Warning!  Could not relabel %s with %s, not relabeling.\n"), tty, newdev_context);
-      syslog (LOG_NOTICE, "Warning!  Could not relabel %s with %s, not relabeling.\n", tty, newdev_context);
-      return NULL;
-    }
-  freecon (newdev_context);
-  return prev_context;
-}
-
-static void
-security_restorelabel_tty (const char *tty, security_context_t context)
-{
-  if (context == NULL)
-    return;
-
-  if (setfilecon (tty, context))
-    {
-      fprintf (stderr, _("Warning!  Could not relabel %s with %s, not relabeling.\n"), tty, context);
-      syslog (LOG_NOTICE, "Warning!  Could not relabel %s with %s, not relabeling.\n", tty, context);
-    }
-  freecon (context);
-  return;
-}
-#endif /* WITH_SELINUX */
-
-
 /* Nice and simple code provided by Linus Torvalds 16-Feb-93 */
 /* Nonblocking stuff by Maciej W. Rozycki, macro@ds2.pg.gda.pl, 1999.
    He writes: "Login performs open() on a tty in a blocking mode.
@@ -323,123 +293,18 @@ check_ttyname (char *ttyn)
     }
 }
 
-/* true if the filedescriptor fd is a console tty, very Linux specific */
-/* XXX Does this work with serial console ?? (kukuk@suse.de) */
-static int
-consoletty (int fd)
-{
-  struct stat stb;
-
-  if ((fstat (fd, &stb) >= 0)
-      && (major (stb.st_rdev) == TTY_MAJOR) && (minor (stb.st_rdev) < 64))
-    return 1;
-  return 0;
-}
-
-/* Create btmp entry if configured for it. */
-static void
-logbtmp (const char *line, const char *username, const char *hostname)
-{
-  struct utmp ut;
-  struct timeval ut_tv;
-  char *path_btmp = getdef_str ("FTMP_FILE");
-
-  /* If no FTMP_FILE is defined in /etc/login.defs, we shouldn't
-     add btmp entries */
-  if (path_btmp == NULL)
-    return;
-
-  memset (&ut, 0, sizeof (ut));
-
-  /* We made sure, that logbtmp is only called with a valid
-     username.  */
-  strncpy (ut.ut_user, username, sizeof (ut.ut_user));
-
-  strncpy (ut.ut_id, line + 3, sizeof (ut.ut_id));
-  strncpy (ut.ut_line, line, sizeof (ut.ut_line));
-  ut.ut_line[sizeof (ut.ut_line) - 1] = 0;
-  gettimeofday (&ut_tv, NULL);
-  ut.ut_tv.tv_sec = ut_tv.tv_sec;
-  ut.ut_tv.tv_usec = ut_tv.tv_usec;
-  ut.ut_type = LOGIN_PROCESS;
-  ut.ut_pid = getpid ();
-  if (hostname)
-    {
-      strncpy (ut.ut_host, hostname, sizeof (ut.ut_host));
-      ut.ut_host[sizeof (ut.ut_host) - 1] = 0;
-      if (hostaddress && hostaddress->h_addr_list)
-	memcpy (&ut.ut_addr, hostaddress->h_addr_list[0], sizeof (ut.ut_addr));
-    }
-
-  updwtmp (path_btmp, &ut);
-}
-
-/* If we create a lastlog file, show the data from it. */
-static void
-dolastlog (int quiet, uid_t uid, const char *tty, const char *hostname)
-{
-  struct lastlog ll;
-  time_t ll_time;
-  int fd;
-
-  if (!getdef_bool ("LASTLOG_ENAB"))	/* give last login and log this one */
-    return;
-
-  if ((fd = open (_PATH_LASTLOG, O_RDWR, 0)) >= 0)
-    {
-      if (lseek (fd, (off_t) uid * sizeof (ll), SEEK_SET) == (off_t)-1)
-	{
-	  int errsv = errno;
-	  fprintf (stderr, _("login: lseek() failed: %s\n"),
-		   strerror (errsv));
-	  return;
-	}
-      if (!quiet)
-	{
-	  if (read (fd, (char *) &ll, sizeof (ll)) == sizeof (ll) &&
-	      ll.ll_time != 0)
-	    {
-	      ll_time = ll.ll_time;
-
-	      printf (_("Last login: %.*s "),
-		      24 - 5, ctime (&ll_time));
-
-	      if (*ll.ll_host != '\0')
-		printf (_("from %.*s\n"),
-			(int) sizeof (ll.ll_host), ll.ll_host);
-	      else
-		printf (_("on %.*s\n"),
-			(int) sizeof (ll.ll_line), ll.ll_line);
-	    }
-	  lseek (fd, (off_t) uid * sizeof (ll), SEEK_SET);
-	}
-      memset (&ll, 0, sizeof (ll));
-      time (&ll_time);
-      ll.ll_time = ll_time;
-      strncpy (ll.ll_line, tty, sizeof (ll.ll_line));
-      ll.ll_line[sizeof (ll.ll_line) - 1] = 0;
-      if (hostname)
-	{
-	  strncpy (ll.ll_host, hostname, sizeof (ll.ll_host));
-	  ll.ll_host[sizeof (ll.ll_host) - 1] = 0;
-	}
-      write (fd, &ll, sizeof (ll));
-      close (fd);
-    }
-}
-
 /* Find out which TERM variable we need to set for this terminal. */
 static char *
-ttytype (const char *line)
+search_ttytype (const char *line)
 {
   FILE *fp;
   char buf[BUFSIZ];
-  char *typefile;
+  const char *typefile;
   char *cp;
   char type[BUFSIZ];
   char port[BUFSIZ];
 
-  if ((typefile = getdef_str ("TTYTYPE_FILE")) == NULL)
+  if ((typefile = getlogindefs_str ("TTYTYPE_FILE", NULL)) == NULL)
     return strdup ("dump");
 
   if (access (typefile, F_OK))
@@ -479,8 +344,11 @@ new_pam_prompt (char *thishost)
   char *ret;
 
   *cp++ = '\n';
-  cp = stpcpy (cp, thishost);
-  cp = stpcpy (cp, " login: ");
+  if (thishost[0]) {
+	cp = stpcpy (cp, thishost);
+	*cp++ = ' ';
+  }
+  cp = stpcpy (cp, "login: ");
 
   ret = strdup (buffer);
 
@@ -494,23 +362,55 @@ new_pam_prompt (char *thishost)
 static inline void
 init_sighandler (void)
 {
-  timeout = getdef_num ("LOGIN_TIMEOUT", 60);
+  timeout = getlogindefs_num ("LOGIN_TIMEOUT", 60);
 
   signal (SIGALRM, timedout);
   alarm ((unsigned int) timeout);
   signal (SIGQUIT, SIG_IGN);
   signal (SIGINT, SIG_IGN);
+  /* We need SIG Child for waitpid */
+  signal (SIGCHLD, SIG_DFL);
 }
 
+#ifdef WITH_AUDIT
+static void
+logaudit (const char *tty, const char *username,
+	  const char *hostname, struct passwd *pwd, int status)
+{
+  const char *userptr;
+  int audit_fd;
+
+  audit_fd = audit_open ();
+  if (audit_fd == -1)
+    return;
+
+  if (pwd && pwd->pw_name)
+    userptr = pwd->pw_name;
+  else if (username)
+    userptr = username;
+  else
+    userptr = "(unknown)";
+
+  audit_log_acct_message (audit_fd, AUDIT_USER_LOGIN,
+			  NULL, "login", userptr, -1,
+			  hostname,  NULL, tty, status);
+
+  close (audit_fd);
+}
+#else /* ! HAVE_LIBAUDIT */
+# define logaudit(tty, username, hostname, pwd, status)
+#endif /* HAVE_LIBAUDIT */
+
 int
 main (int argc, char **argv)
 {
   struct group *gr;
-  int ask, fflag, hflag, pflag, cnt, errsv;
+  int fflag, hflag, pflag, cnt, errsv;
   int quietlog, passwd_req;
-  char *domain, *ttyn, *tty, *group;
-  char thishost[MAXPATHLEN + 1];
-  char *termenv;
+  char *domain, *ttyn, *tty;
+  const char *group;
+  char thishost[MAXHOSTNAMELEN + 1];
+  char *termenv = NULL;
   char *childArgv[10];
   char *buff;
   int childArgc = 0;
@@ -518,36 +418,58 @@ main (int argc, char **argv)
   pam_handle_t *pamh = NULL;
   struct pam_conv conv = { misc_conv, NULL };
   pid_t childPid;
-  char *hushfile;
+  const char *hushfile;
   gid_t gid;
-  struct faillog faillog;
   char *hostname;
   const char *username;
   const void *void_username;
-  int retries = getdef_num ("LOGIN_RETRIES", 3);
+  int retries = getlogindefs_num ("LOGIN_RETRIES", 3);
   int buflen = 256;
   char *buffer = alloca (buflen);
   struct passwd resultbuf;
   struct passwd *pwd;
-  char *vcsn = NULL, *vcsan = NULL;
-
+  int tty_perm = getlogindefs_num ("TTYPERM", 0600);
+  const char *locale;
 
   init_sighandler ();
 
-  setlocale (LC_ALL, "");
-  bindtextdomain ("pam_login", LOCALEDIR);
-  textdomain ("pam_login");
+  /* find locale for translations:
+     If the locale environment variables are already set
+     set when the login program is called, just use them.
+     Usually they are not set though, because
+     - this program is called before the user has logged
+       in, e.g. when logging in from the Linux console
+       or from a serial console and no profiles have
+       sourced yet.
+     - this program is called by the telnet server, in
+       that case only the environment variables
+       DISPLAY, TERM, and REMOTEHOST are set.
+     If the locale environment is not set, determine the
+     language from /etc/sysconfig/language.                  */
+
+  if ((locale = getenv ("LC_ALL")) == NULL
+      || strlen (locale) == 0)
+    if ((locale = getenv ("LC_MESSAGES")) == NULL
+	|| strlen (locale) == 0)
+      if ((locale = getenv ("LANG")) == NULL
+	  || strlen (locale) == 0)
+        if ((locale = getlogindefs_str ("RC_LC_ALL", NULL)) == NULL
+	    || strlen (locale) == 0)
+          if ((locale = getlogindefs_str ("RC_LC_MESSAGES", NULL)) == NULL
+	      || strlen (locale) == 0)
+            locale = getlogindefs_str ("RC_LANG", "POSIX");
+
+  setlocale (LC_ALL, locale);
+  bindtextdomain ("login", LOCALEDIR);
+  textdomain ("login");
 
   setpriority (PRIO_PROCESS, 0, 0);
-#ifdef HAVE_QUOTA
-  quota (Q_SETUID, 0, 0, 0);
-#endif
   initproctitle (argc, argv);
 
   if (gethostname (thishost, sizeof (thishost)) < 0)
     {
       errsv = errno;
-      fprintf (stderr, _("login: gethostname() failed: %s\n"),
+      fprintf (stderr, "login: gethostname() failed: %s\n",
 	       strerror (errsv));
       strncpy (thishost, "localhost", sizeof (thishost));
     }
@@ -558,7 +480,6 @@ main (int argc, char **argv)
   fflag = hflag = pflag = 0;
   passwd_req = 1;
 
-
   while (1)
     {
       int c;
@@ -570,13 +491,14 @@ main (int argc, char **argv)
 	{NULL, 0, NULL, '\0'}
       };
 
-      c = getopt_long (argc, argv, "fh:p", long_options, &option_index);
+      c = getopt_long (argc, argv, "f:h:pH", long_options, &option_index);
       if (c == (-1))
 	break;
       switch (c)
 	{
 	case 'f':
 	  fflag = 1;
+	  username = strdup (optarg);
 	  break;
 
 	case 'h':
@@ -584,7 +506,6 @@ main (int argc, char **argv)
 	    char *cp;
 	    int hostbuflen = 512;
 	    char *hostbuffer = malloc (hostbuflen);
-	    struct hostent resulthostbuf;
 
 	    if (getuid ())
 	      {
@@ -594,7 +515,7 @@ main (int argc, char **argv)
 
 	    if (hostbuffer == NULL)
 	      {
-		fprintf (stderr, _("login: out of memory.\n"));
+		fprintf (stderr, "login: out of memory.\n");
 		exit (1);
 	      }
 
@@ -606,7 +527,7 @@ main (int argc, char **argv)
 	    hostname = strdup (optarg);
 	    if (hostname == NULL)
 	      {
-		fprintf (stderr, _("login: out of memory.\n"));
+		fprintf (stderr, "login: out of memory.\n");
 		exit (1);
 	      }
 
@@ -625,7 +546,7 @@ main (int argc, char **argv)
 		  hostbuffer = buf;
 		else
 		  {
-		    fprintf (stderr, _("login: out of memory.\n"));
+		    fprintf (stderr, "login: out of memory.\n");
 		    if (hostbuffer)
 		      free (hostbuffer);
 		    exit (1);
@@ -637,6 +558,9 @@ main (int argc, char **argv)
 	case 'p':
 	  pflag = 1;
 	  break;
+	case 'H':
+	  thishost[0] = '\0';
+	  break;
 	case '\253':
 	  print_help ();
 	  return 0;
@@ -658,7 +582,7 @@ main (int argc, char **argv)
   if (!isatty(0) || !isatty(1) || !isatty(2))
     return 1;                /* must be a terminal */
 
-  if (argc > 1)
+  if (argc > 1 || (argc >= 1 && fflag))
     {
       fputs (_("login: Too many arguments\n"), stderr);
       print_error ();
@@ -668,20 +592,34 @@ main (int argc, char **argv)
   if (*argv)
     {
       char *cp = *argv;
+
       username = strdup (cp);
       if (username == NULL)
 	{
-	  fputs (_("login: out of memory.\n"), stderr);
+	  fputs ("login: out of memory.\n", stderr);
 	  exit (1);
 	}
-      ask = 0;
       /* wipe name - some people mistype their password here */
       /* (of course we are too late, but perhaps this helps a little ..) */
       while (*cp)
 	*cp++ = ' ';
     }
-  else
-    ask = 1;
+
+#ifdef WITH_DEBUG
+  if (username != NULL && username[1] == '@')
+    {
+      if (username[0] == '+')
+	{
+	  do_debug = 1;
+	  username +=2;
+	}
+      else if (username[0] == '@')
+	{
+	  do_debug = 2;
+	  username += 2;
+	}
+    }
+#endif
 
   for (cnt = getdtablesize (); cnt > 2; cnt--)
     close (cnt);
@@ -690,7 +628,15 @@ main (int argc, char **argv)
 
   openlog ("login", LOG_ODELAY | LOG_PID, LOG_AUTHPRIV);
 
+#ifdef WITH_DEBUG
+  if (username)
+    debug ("username = '%s'", username);
+#endif
+
   ttyn = ttyname (0);
+#ifdef WITH_DEBUG
+  debug ("ttyname = '%s'", ttyn);
+#endif
   if (ttyn != NULL && *ttyn != '\0')
     {
       struct termios tt, ttt;
@@ -703,15 +649,21 @@ main (int argc, char **argv)
       ttt.c_cflag &= ~HUPCL;
 
       if ((chown (ttyn, 0, 0) == -1) ||
-	  (chmod (ttyn, getdef_num ("TTYPERM", 0600)) == -1))
+	  (chmod (ttyn, tty_perm) == -1))
 	{
 	  errsv = errno;
+
 	  syslog (LOG_ERR, "FATAL: cannot change permissions of TTY: %s",
 		  strerror (errsv));
-	  fprintf (stderr, _("FATAL: cannot change permissions of TTY: %s"),
+	  fprintf (stderr,
+		   _("FATAL: cannot change permissions of TTY: %s"),
 		   strerror (errsv));
 	  sleep (5);
-	  exit (1);
+
+	  /* don't exit if file system is readonly, or root cannot
+	     login and fix it. */
+	  if (errsv != EROFS)
+	    exit (1);
 	}
 
       tcsetattr (0, TCSAFLUSH, &ttt);
@@ -741,7 +693,15 @@ main (int argc, char **argv)
      Therefore, we are safe not setting it to anything
   */
 
+#ifdef WITH_DEBUG
+  debug ("before pam_start");
+#endif
+
+#if !defined(WITHOUT_REMOTE)
+  retcode = pam_start (hflag?"remote":"login", username, &conv, &pamh);
+#else
   retcode = pam_start ("login", username, &conv, &pamh);
+#endif
   if (retcode != PAM_SUCCESS)
     {
       fprintf (stderr, _("login: PAM Failure, aborting: %s\n"),
@@ -752,13 +712,22 @@ main (int argc, char **argv)
     }
   /* hostname & tty are either set to NULL or their correct values,
      depending on how much we know */
+#ifdef WITH_DEBUG
+  debug ("pam_set_item (pamh, PAM_RHOST, %s)", hostname?hostname:"NULL");
+#endif
   retcode = pam_set_item (pamh, PAM_RHOST, hostname);
   PAM_FAIL_CHECK;
   if (ttyn)
     {
+#ifdef WITH_DEBUG
+      debug ("pam_set_item (pamh, PAM_TTY, %s)", ttyn);
+#endif
       retcode = pam_set_item (pamh, PAM_TTY, ttyn);
       PAM_FAIL_CHECK;
     }
+#ifdef WITH_DEBUG
+  debug ("pam_set_item (pamh, PAM_USER_PROMPT, %s)", new_pam_prompt (thishost));
+#endif
   retcode = pam_set_item (pamh, PAM_USER_PROMPT, new_pam_prompt (thishost));
   PAM_FAIL_CHECK;
 
@@ -779,8 +748,13 @@ main (int argc, char **argv)
          pay attention to failure count and get rid of MAX_LOGIN_TRIES? */
 
       do {
+#ifdef WITH_DEBUG
+	debug ("before pam_authenticate ()");
+#endif
 	retcode = pam_authenticate (pamh, 0);
-
+#ifdef WITH_DEBUG
+	debug ("pam_authenticate returned %d", retcode);
+#endif
 	if (retcode != PAM_SUCCESS)
 	  {
 	    const char *userptr;
@@ -800,7 +774,7 @@ main (int argc, char **argv)
 
 	    if (pwd)
 	      userptr = pwd->pw_name;
-	    else if (getdef_bool ("LOG_UNKFAIL_ENAB"))
+	    else if (getlogindefs_bool ("LOG_UNKFAIL_ENAB", 0))
 	      userptr = username;
 	    else
 	      userptr = "UNKNOWN";
@@ -814,18 +788,12 @@ main (int argc, char **argv)
 		      failcount, hostname ? hostname :
 		      (ttyn ? ttyn : "<unknown>"), userptr,
 		      pam_strerror (pamh, retcode));
-	    if (ttyn && strlen (ttyn) > strlen (_PATH_DEV))
-	      logbtmp (ttyn + strlen (_PATH_DEV), userptr, hostname);
-	    else
-	      syslog (LOG_NOTICE, "btmp logging disabled, no tty found");
-	    if (getdef_bool ("FAILLOG_ENAB"))
-	      {
-		if (pwd)
-		  failure (pwd->pw_uid, tty, &faillog);
-	      }
+
+	    logaudit (tty, username, hostname, pwd, 0);
+
 	    fprintf (stderr, _("Login incorrect\n\n"));
 	    pam_set_item (pamh, PAM_USER, NULL);
-	    sleep (getdef_num ("FAIL_DELAY", 1));
+	    sleep (getlogindefs_num ("FAIL_DELAY", 1));
 	  }
       } while ((failcount < retries) &&
 	       ((retcode == PAM_AUTH_ERR) ||
@@ -838,14 +806,30 @@ main (int argc, char **argv)
 	  pam_end (pamh, retcode);
 	  exit (0);
 	}
+#ifdef WITH_AUDIT
+      else
+	{
+	  pam_get_item (pamh, PAM_USER, &void_username);
+	  username = void_username;
+	  if (!*username)
+	    username = NULL;
 
+	  logaudit (tty, username, hostname, NULL, 1);
+	}
+#endif
+    }
+
+#ifdef WITH_DEBUG
+      debug ("before pam_acct_mgmt");
+#endif
       retcode = pam_acct_mgmt (pamh, 0);
-
+#ifdef WITH_DEBUG
+      debug ("pam_acct_mgmt returned %d", retcode);
+#endif
       if (retcode == PAM_NEW_AUTHTOK_REQD)
 	retcode = pam_chauthtok (pamh, PAM_CHANGE_EXPIRED_AUTHTOK);
 
       PAM_FAIL_CHECK;
-    }
 
   /* Grab the user information out of the password file for future usage
    * First get the username that we are actually using, though.
@@ -878,14 +862,6 @@ main (int argc, char **argv)
       exit (1);
     }
 
-  if (getdef_bool ("FAILLOG_ENAB") && !failcheck (pwd->pw_uid))
-    {
-      printf (_("exceeded failure limit for `%s'%s\n"), pwd->pw_name, hostname);
-      syslog (LOG_CRIT, "exceeded failure limit for `%s'%s\n",
-	      pwd->pw_name, hostname);
-      exit (1);
-    }
-
   /* Initialize the supplementary group list. This should be done
      before pam_setcred because the PAM modules might add groups
      during pam_setcred. For root we don't call initgroups, instead
@@ -914,87 +890,27 @@ main (int argc, char **argv)
       exit (1);
     }
 
+#ifdef WITH_DEBUG
+  debug ("pam_setcred (pamh, PAM_ESTABLISH_CRED)");
+#endif
   retcode = pam_setcred (pamh, PAM_ESTABLISH_CRED);
   PAM_FAIL_CHECK;
 
+#ifdef WITH_DEBUG
+  debug ("pam_open_session (pamh, 0)");
+#endif
   retcode = pam_open_session (pamh, 0);
   PAM_FAIL_CHECK;
 
-#ifdef WITH_SELINUX
-  /* Make sure SELINUX is really running on this system */
-  selinux_enabled = is_selinux_enabled ();
-  if (selinux_enabled > 0)
-    {
-      int ret;
-      security_context_t* contextlist;
-      int num_contexts = get_ordered_context_list (username, 0, &contextlist);
-      if (num_contexts > 0)
-	{
-#if 1
-	  /* Allow the user to select the context, if more than
-	     one exist.  */
-	  ret = query_user_context (contextlist, &user_context);
-	  freeconary (contextlist);
-
-	  if (ret < 0)
-	    {
-	      fprintf (stderr, _("%s:  query_user_context failed\n"), argv[0]);
-	      syslog (LOG_ERR, "%s:  query_user_context failed\n", argv[0]);
-	      sleep (5);
-	      exit (3);
-	    }
-#else
-	  /* Always use the default context. User should use "newrole"
-	     to change the context.  */
-	  user_context = (security_context_t) strdup (contextlist[0]);
-	  freeconary (contextlist);
-#endif
-	}
-      else
-	{
-	  fprintf (stderr, _("login: unable to obtain context for %s.\n"),
-		   username);
-	  ret = manual_user_enter_context (username, &user_context);
-	  if (ret < 0)
-	    {
-	      syslog (LOG_ERR, "UNABLE TO GET VALID CONTEXT FOR %s",
-		      username);
-	      sleep (5);
-	      exit (3);
-	    }
-	}
-    }
-#endif
-
   /* committed to login -- turn off timeout */
   alarm ((unsigned int) 0);
 
-#ifdef HAVE_QUOTA
-  if (quota (Q_SETUID, pwd->pw_uid, 0, 0) < 0 && errno != EINVAL)
-    {
-      switch (errno)
-	{
-	case EUSERS:
-	  fprintf (stderr,
-		   _("Too many users logged on already.\nTry again later.\n"));
-	  break;
-	case EPROCLIM:
-	  fprintf (stderr,
-		   _("You have too many processes running.\n"));
-	  break;
-	default:
-	  perror ("quota (Q_SETUID)");
-	}
-      sleep (5);
-      exit (0);           /* %% */
-    }
-#endif
-
   /* paranoia... */
   endpwent ();
 
   quietlog = 0;
-  if ((hushfile = getdef_str ("HUSHLOGIN_FILE")) != NULL)
+  hushfile = getlogindefs_str ("HUSHLOGIN_FILE", NULL);
+  if (hushfile != NULL && hushfile[0] != '\0')
     {
       /* If this is not a fully rooted path then see if the
          file exists in the user's home directory. */
@@ -1067,7 +983,7 @@ main (int argc, char **argv)
      * If we can't find a pre-existing entry by pid, try by line
      * BSD network daemons may rely on this.
      */
-    if (utp == NULL)
+    if (utp == NULL && ttyn != NULL)
       {
 	endutent ();
 	setutent ();
@@ -1083,7 +999,7 @@ main (int argc, char **argv)
     else	    /* some gettys/telnetds don't initialize utmp... */
       memset (&ut, 0, sizeof (ut));
 
-    if (ut.ut_id[0] == 0)
+    if (ut.ut_id[0] == 0 && ttyn != NULL)
       strncpy (ut.ut_id, ttyn + strlen (_PATH_TTY), sizeof (ut.ut_id));
 
     strncpy (ut.ut_user, username, sizeof (ut.ut_user));
@@ -1109,9 +1025,7 @@ main (int argc, char **argv)
     updwtmp (_PATH_WTMP, &ut);
   }
 
-  dolastlog (quietlog, pwd->pw_uid, tty, hostname);  /* Maybe we move this to PAM ? */
-
-  if ((group = getdef_str ("TTYGROUP")) == NULL)
+  if ((group = getlogindefs_str ("TTYGROUP", NULL)) == NULL)
     gid = pwd->pw_gid;
   else if (group[0] >= '0' && group[0] <= '9')
     gid = atoi (group);
@@ -1120,64 +1034,30 @@ main (int argc, char **argv)
   else
     gid = pwd->pw_gid;
 
-  chown (ttyn, pwd->pw_uid, gid);
-  chmod (ttyn,  getdef_num ("TTYPERM", 0600));
-
-#ifdef WITH_SELINUX
-  if (selinux_enabled > 0)
-    ttyn_context = security_label_tty (ttyn, user_context);
-#endif
-
-  /* if tty is one of the VC's then change owner and mode of the
-     special /dev/vcs devices as well */
-  if (consoletty (0))
+  if (ttyn != NULL)
     {
-      /* find names of Virtual Console devices */
-      char *p = ttyn;
-      /* find number of tty */
-      while (*p && !isdigit (*p))
-	++p;
-
-      if (asprintf (&vcsn, "/dev/vcs%s", p) < 0)
-	vcsn = NULL;
-      if (asprintf (&vcsan, "/dev/vcsa%s", p) < 0)
-	vcsan = NULL;
-#if 0
-      /*
-       * Please don't add code to chown /dev/vcs* to the user logging in -
-       * it's a potential security hole.  I wouldn't like the previous user
-       * to hold the file descriptor open and watch my screen.  We don't
-       * have the *BSD revoke() system call yet, and vhangup() only works
-       * for tty devices (which vcs* is not).
-       */
-      if (vcsn)
+      if (chown (ttyn, pwd->pw_uid, gid) < 0)
 	{
-	  chown (vcsn, pwd->pw_uid, gid);
-	  chmod (vcsn, getdef_num ("TTYPERM", 0600));
+	  errsv = errno;
+	  syslog (LOG_ERR, "chown (%s, %lu, %lu) failed: %s",
+		  ttyn, (unsigned long) pwd->pw_uid, (unsigned long) gid,
+		  strerror (errsv));
+	  fprintf (stderr, "login: chown (%s, %lu, %lu) failed: %s\n",
+		   ttyn, (unsigned long) pwd->pw_uid, (unsigned long) gid,
+		   strerror (errsv));
 	}
-      if (vcsan)
+      if (chmod (ttyn, tty_perm) < 0)
 	{
-	  chown (vcsan, pwd->pw_uid, gid);
-	  chmod (vcsan, getdef_num ("TTYPERM", 0600));
+	  errsv = errno;
+	  syslog (LOG_ERR, "chmod (%s, %u) failed: %s\n",
+		  ttyn, tty_perm, strerror (errsv));
+	  fprintf (stderr, "login: chmod (%s, %u) failed: %s\n",
+		   ttyn, tty_perm, strerror (errsv));
 	}
-#endif
-#ifdef WITH_SELINUX
-      if (selinux_enabled > 0)
-        {
-	  if (vcsn)
-	    vcsn_context = security_label_tty (vcsn, user_context);
-	  if (vcsan)
-	    vcsan_context = security_label_tty (vcsan, user_context);
-	}
-#endif
     }
 
   setgid (pwd->pw_gid);
 
-#ifdef HAVE_QUOTA
-  quota (Q_DOWARN, pwd->pw_uid, (dev_t)-1, 0);
-#endif
-
   if (*pwd->pw_shell == '\0')
     pwd->pw_shell = _PATH_BSHELL;
 
@@ -1186,7 +1066,8 @@ main (int argc, char **argv)
     char *ep;
 
     if (!((ep = getenv ("TERM")) && (termenv = strdup (ep))))
-      termenv = ttytype (tty);
+      if (tty)
+	termenv = search_ttytype (tty);
   }
 
   /* destroy environment unless user has requested preservation */
@@ -1197,39 +1078,49 @@ main (int argc, char **argv)
     }
 
   if (setenv ("HOME", pwd->pw_dir, 0) == -1)	/* legal to override */
-    fputs (_("login: setenv (\"HOME\") failed: insufficient space."),
+    fputs ("login: setenv (\"HOME\") failed: insufficient space.",
 	   stderr);
   if (pwd->pw_uid)
-    setenv ("PATH", getdef_str ("ENV_PATH"), 1);
+    setenv ("PATH", getlogindefs_str ("ENV_PATH",
+				      "/usr/local/bin:/usr/bin:/bin"), 1);
   else
-    setenv ("PATH", getdef_str ("ENV_ROOTPATH"), 1);
+    setenv ("PATH", getlogindefs_str ("ENV_ROOTPATH",
+				      "/sbin:/bin:/usr/sbin:/usr/bin"), 1);
 
   if (setenv ("SHELL", pwd->pw_shell, 1) == -1)
-    fputs (_("login: setenv (\"SHELL\") failed: insufficient space."),
-	   stderr);
-  if (setenv ("TERM", termenv, 1) == -1)
-    fputs (_("login: setenv (\"TERM\") failed: insufficient space."),
+    fputs ("login: setenv (\"SHELL\") failed: insufficient space.",
 	   stderr);
+  if (termenv)
+    if (setenv ("TERM", termenv, 1) == -1)
+      fputs ("login: setenv (\"TERM\") failed: insufficient space.",
+	     stderr);
 
   /* mailx will give a funny error msg if you forget this one */
   /* avoid snprintf */
-  if (sizeof (_PATH_MAILDIR) + strlen (pwd->pw_name) + 1 < MAXPATHLEN)
-    {
-      char tmp[MAXPATHLEN];
+  {
+    char *tmp;
 
-      sprintf (tmp, "%s/%s", _PATH_MAILDIR, pwd->pw_name);
-      if (setenv ("MAIL", tmp, 0) == -1)
-	fputs (_("login: setenv (\"MAIL\") failed: insufficient space."),
-	       stderr);
-    }
+    if (asprintf (&tmp, "%s/%s", _PATH_MAILDIR, pwd->pw_name) < 0)
+      {
+	fprintf (stderr, "login: out of memory.\n");
+	exit (1);
+      }
+    else
+      {
+	if (setenv ("MAIL", tmp, 0) == -1)
+	  fputs ("login: setenv (\"MAIL\") failed: insufficient space.",
+		 stderr);
+	free (tmp);
+      }
+  }
 
   /* Export the user name.  For BSD derived systems, it's "USER", for
      all others it's "LOGNAME".  We set both of them. */
   if (setenv ("USER", pwd->pw_name, 1) == -1)
-    fputs (_("login: setenv (\"USER\") failed: insufficient space."),
+    fputs ("login: setenv (\"USER\") failed: insufficient space.",
 	   stderr);
   if (setenv ("LOGNAME", pwd->pw_name, 1) == -1)
-    fputs (_("login: setenv (\"LOGNAME\") failed: insufficient space."),
+    fputs ("login: setenv (\"LOGNAME\") failed: insufficient space.",
 	   stderr);
 
   {
@@ -1250,30 +1141,6 @@ main (int argc, char **argv)
 
   setproctitle ("login", username);
 
-#ifdef WITH_SELINUX
-  if (selinux_enabled > 0)
-    {
-      if (pwd->pw_uid == 0)
-	{
-	  if (hostname)
-	    syslog (LOG_NOTICE, "ROOT LOGIN ON %s FROM %s USING %s",
-		    ttyn, hostname, user_context);
-	  else
-	    syslog (LOG_NOTICE, "ROOT LOGIN ON %s USING %s", ttyn,
-		    user_context);
-	}
-      else
-	{
-	  if (hostname)
-	    syslog (LOG_INFO, "LOGIN ON %s BY %s FROM %s USING %s",
-		    ttyn, pwd->pw_name, hostname, user_context);
-	  else
-	    syslog (LOG_INFO, "LOGIN ON %s BY %s USING %s", ttyn,
-		    pwd->pw_name, user_context);
-	}
-    }
-#endif
-
   if (!quietlog)
     motd ();
 
@@ -1294,19 +1161,23 @@ main (int argc, char **argv)
   signal (SIGHUP, SIG_IGN);
   ioctl(0, TIOCNOTTY, NULL);
   signal (SIGHUP, SIG_DFL);
+#ifdef WITH_DEBUG
+  debug ("fork child process");
+#endif
   childPid = fork ();
   if (childPid < 0)
     {
       errsv = errno;
       /* error in fork() */
-      fprintf (stderr, _("login: failure forking: %s"), strerror (errsv));
+      fprintf (stderr, "login: failure forking: %s", strerror (errsv));
       PAM_END;
       exit (0);
     }
   else if (childPid)
     {
       /* parent - wait for child to finish, then cleanup session */
-      chdir ("/");
+      if (chdir ("/") == -1)
+	{ /* ignore */ };
 
       signal(SIGINT, SIG_IGN);
       signal(SIGQUIT, SIG_IGN);
@@ -1318,28 +1189,12 @@ main (int argc, char **argv)
 	{
 	  errsv = errno;
 	  if (errsv != EINTR)
-	    fprintf (stderr, _("login: waitpid (%d, NULL, 0) failed: %s\n"),
-		     childPid, strerror (errsv));
-	}
-#ifdef WITH_SELINUX
-      if (selinux_enabled > 0)
-        {
-          /* We need to change the contexts of the terminal devices back to
-             the system when the user's session ends.  */
-          security_restorelabel_tty (ttyn, ttyn_context);
-          if (consoletty (0))
 	    {
-	      if (vcsn)
-		security_restorelabel_tty (vcsn, vcsn_context);
-	      if (vcsan)
-		security_restorelabel_tty (vcsan, vcsan_context);
+	      fprintf (stderr, "login: waitpid (%d, NULL, 0) failed: %s\n",
+		       childPid, strerror (errsv));
+	      break;
 	    }
-        }
-#endif
-      if (vcsn)
-	free (vcsn);
-      if (vcsan)
-	free (vcsan);
+	}
 
       PAM_END;
       exit (0);
@@ -1350,7 +1205,7 @@ main (int argc, char **argv)
   if (ioctl (0, TIOCSCTTY, (char *) 1) == -1)
     {
       errsv = errno;
-      fprintf (stderr, _("login: couldn't set controlling tty: %s\n"),
+      fprintf (stderr, "login: couldn't set controlling tty: %s\n",
 	       strerror (errsv));
       exit (1);
     }
@@ -1359,22 +1214,6 @@ main (int argc, char **argv)
 
   openlog ("login", LOG_ODELAY | LOG_PID, LOG_AUTHPRIV);
 
-#ifdef WITH_SELINUX
-  if (selinux_enabled > 0)
-    {
-      if (setexeccon (user_context))
-	{
-	  fprintf (stderr,
-		   _("Error!  Unable to set executable context %s.\n"),
-		   user_context);
-	  syslog(LOG_ERR, "Error!  Unable to set executable context %s.\n",
-		 user_context);
-	  exit (3);
-	}
-      freecon (user_context);
-    }
-#endif
-
   /* discard permissions last so can't get killed and drop core */
   if (setuid (pwd->pw_uid) < 0 && pwd->pw_uid)
     {
@@ -1386,7 +1225,7 @@ main (int argc, char **argv)
   if (chdir (pwd->pw_dir) < 0)
     {
       printf (_("No directory %s!\n"), pwd->pw_dir);
-      if (!getdef_bool ("DEFAULT_HOME"))
+      if (!getlogindefs_bool ("DEFAULT_HOME", 1))
 	exit (0);
       if (chdir ("/"))
 	exit (0);
@@ -1401,7 +1240,7 @@ main (int argc, char **argv)
 
       if (!buff)
 	{
-	  fprintf (stderr, _("login: no memory for shell script.\n"));
+	  fprintf (stderr, "login: no memory for shell script.\n");
 	  exit (0);
 	}
 
@@ -1428,6 +1267,9 @@ main (int argc, char **argv)
 
   childArgv[childArgc++] = NULL;
 
+#ifdef WITH_DEBUG
+  debug ("Calling execvp (%s)", childArgv[0]);
+#endif
   execvp (childArgv[0], childArgv + 1);
 
   errsv = errno;
